with Grt.Types; use Grt.Types;
with Grt.Signals; use Grt.Signals;
with Grt.Processes;
with Grt.Hierarchy; use Grt.Hierarchy;
with Ada.Characters.Latin_1;
with System; use System;
with Grt.Errors; use Grt.Errors;

package body Grt.Disp is
   procedure Put (Stream : FILEs; Str : String)
   is
      S : size_t;
   begin
      S := fwrite (Str'Address, Str'Length, 1, Stream);
   end Put;

   procedure Put (Str : String)
   is
      S : size_t;
   begin
      S := fwrite (Str'Address, Str'Length, 1, stdout);
   end Put;

   procedure Put (C : Character)
   is
      R : int;
   begin
      R := fputc (Character'Pos (C), stdout);
   end Put;

   procedure New_Line is
   begin
      Put (Ada.Characters.Latin_1.LF);
   end New_Line;

   procedure Put_Line (Str : String)
   is
   begin
      Put (Str);
      New_Line;
   end Put_Line;

--    function Get_String (Str : Ghdl_Str_Len_Type) return String
--    is
--    begin
--       if Str.Str = null then
--          return ''' & Character'Val (Str.Len) & ''';
--       else
--          return Str.Str (1 .. Str.Len);
--       end if;
--    end Get_String;

   procedure Put_Str_Len (Stream : FILEs; Str : Ghdl_Str_Len_Type)
   is
      S : String (1 .. 3);
   begin
      if Str.Str = null then
         S (1) := ''';
         S (2) := Character'Val (Str.Len);
         S (3) := ''';
         Put (Stream, S);
      else
         Put (Stream, Str.Str (1 .. Str.Len));
      end if;
   end Put_Str_Len;

--    procedure Put_Trim (Stream : FILEs; Str : String)
--    is
--       Start : Natural;
--    begin
--       Start := Str'First;
--       while Start <= Str'Last and then Str (Start) = ' ' loop
--          Start := Start + 1;
--       end loop;
--       Put (Stream, Str (Start .. Str'Last));
--    end Put_Trim;

   generic
      type Ntype is range <>;
      Max_Len : Natural;
   procedure Put_Ntype (Stream : FILEs; N : Ntype);

   procedure Put_Ntype (Stream : FILEs; N : Ntype)
   is
      Str : String (1 .. Max_Len);
      P : Natural := Str'Last;
      V : Ntype;
   begin
      if N > 0 then
         V := -N;
      else
         V := N;
      end if;
      loop
         Str (P) := Character'Val (48 - (V rem 10));
         V := V / 10;
         exit when V = 0;
         P := P - 1;
      end loop;
      if N < 0 then
         P := P - 1;
         Str (P) := '-';
      end if;
      Put (Stream, Str (P .. Max_Len));
   end Put_Ntype;

   procedure Put_I32_1 is new Put_Ntype (Ntype => Ghdl_I32, Max_Len => 11);

   procedure Put_I32 (Stream : FILEs; I32 : Ghdl_I32) renames Put_I32_1;

   procedure Put_P64_1 is new Put_Ntype (Ntype => Ghdl_P64, Max_Len => 19);
   procedure Put_P64 (Stream : FILEs; P64 : Ghdl_P64) renames Put_P64_1;

   procedure Put_F64 (Stream : FILEs; F64 : Ghdl_F64)
   is
      procedure fprintf (Stream : FILEs;
                         Template : System.Address;
                         Arg : Ghdl_F64);
      pragma Import (C, fprintf);

      Str : constant String := "%g" & Character'Val (0);
   begin
      fprintf (Stream, Str'Address, F64);
   end Put_F64;

   procedure Put_E8 (Stream : FILEs; E8 : Ghdl_E8; Type_Desc : Ghdl_Desc_Ptr)
   is
   begin
      Put_Str_Len (Stream, Type_Desc.E8.Values (Natural (E8)));
   end Put_E8;

   --procedure Put_E32
   --  (Stream : FILEs; E32 : Ghdl_E32; Type_Desc : Ghdl_Desc_Ptr)
   --is
   --begin
   --   Put_Str_Len (Stream, Type_Desc.E32.Values (Natural (E32)));
   --end Put_E32;

   procedure Put_Dir (Dir : Ghdl_Dir_Type) is
   begin
      case Dir is
         when Dir_To =>
            Put ("to");
         when Dir_Downto =>
            Put ("downto");
      end case;
   end Put_Dir;

   procedure Put_Sig_Index (Sig : Sig_Table_Index)
   is
   begin
      Put_I32 (stdout, Ghdl_I32 (Sig));
   end Put_Sig_Index;

   procedure Put_Sig_Range (Sig : Sig_Table_Range)
   is
   begin
      Put_Sig_Index (Sig.First);
      if Sig.Last /= Sig.First then
         Put ("-");
         Put_Sig_Index (Sig.Last);
      end if;
   end Put_Sig_Range;

   procedure Disp_Value (Stream : FILEs;
                         Value : Value_Union;
                         Type_Desc : Ghdl_Desc_Ptr)
   is
      Td : Ghdl_Desc_Ptr;
   begin
      if Type_Desc.Kind in Ghdl_Subtype_Scalar_Kind then
         Td := Type_Desc.St.Basetype;
      else
         Td := Type_Desc;
      end if;

      case Td.Kind is
         when Ghdl_Type_B2 =>
            case Value.B2 is
               when True =>
                  Put_Str_Len (Stream, Td.B2.True_Name);
               when False =>
                  Put_Str_Len (Stream, Td.B2.False_Name);
            end case;
         when Ghdl_Type_E8 =>
            Put_E8 (Stream, Value.E8, Td);
            --Put (Ghdl_E8'Image (Value.E8));
         when Ghdl_Type_E32 =>
            --Put (Stream, Ghdl_E32'Image (Value.E32));
            Put_I32 (Stream, Ghdl_I32 (Value.E32));
         when Ghdl_Type_I32 =>
            Put_I32 (Stream, Value.I32);
         when Ghdl_Type_P64 =>
            Put_P64 (Stream, Value.P64);
         when Ghdl_Type_F64 =>
            Put_F64 (Stream, Value.F64);
         when others =>
            Put (Stream, "???");
      end case;
   end Disp_Value;

   procedure Disp_Time (T : Std_Time) is
   begin
      Put_P64 (stdout, Ghdl_P64 (T));
   end Disp_Time;

   procedure Disp_Instance_Name (Inst : Ghdl_Instance_Name_Acc)
   is
   begin
      if Inst.Parent /= null then
         Disp_Instance_Name (Inst.Parent);
         Put ('.');
      end if;
      case Inst.Kind is
         when Ghdl_Name_Architecture =>
            Put ('(');
            Put_Str_Len (stdout, Inst.Name.all);
            Put (')');
         when others =>
            if Inst.Name /= null then
               Put_Str_Len (stdout, Inst.Name.all);
            end if;
      end case;
   end Disp_Instance_Name;

--    procedure Disp_Signal_Name (Sig : Ghdl_Signal_Ptr) is
--    begin
--       if Sig.Instance_Name /= null then
--          Disp_Instance_Name (Sig.Instance_Name);
--       else
--          Put ("???");
--       end if;
--    end Disp_Signal_Name;

   procedure Disp_Hierarchy_Instance_Name (Inst : Ghdl_Instance_Name_Acc)
   is
   begin
      Put_Str_Len (stdout, Inst.Name.all);
      case Inst.Kind is
         when Ghdl_Name_Entity =>
            Put (" [entity]");
         when Ghdl_Name_Architecture =>
            Put (" [arch]");
         when Ghdl_Name_Block =>
            Put (" [block]");
         when Ghdl_Name_Generate_Iterative =>
            Put ("(");
            case Inst.Iter_Typedesc.Kind is
               when Ghdl_Type_I32 =>
                  Put_I32 (stdout, Inst.Iter_Addr.I32);
               when Ghdl_Type_E8 =>
                  Put_E8 (stdout, Inst.Iter_Addr.E8, Inst.Iter_Typedesc);
               when others =>
                  Put ("?");
            end case;
            Put (")");
            Put (" [for-generate]");
         when Ghdl_Name_Generate_Conditional =>
            Put (" [if-generate]");
         when Ghdl_Name_Instance =>
            Put (" [instance]");
         when Ghdl_Name_Signal =>
            case Inst.Sig_Mode is
               when Mode_Signal =>
                  Put (" (signal)");
               when Mode_In =>
                  Put (" (in)");
               when Mode_Out =>
                  Put (" (out)");
               when Mode_Inout =>
                  Put (" (inout)");
               when Mode_Buffer =>
                  Put (" (buffer)");
               when Mode_Linkage =>
                  Put (" (linkage)");
               when Mode_Stable =>
                  Put (" ('stable)");
               when Mode_Quiet =>
                  Put (" ('quiet)");
               when Mode_Transaction =>
                  Put (" ('transaction)");
               when Mode_Delayed =>
                  Put (" ('delayed)");
               when Mode_Guard =>
                  Put (" (guard)");
               when Mode_Conv_In =>
                  Put (" (conv_in)");
               when Mode_Conv_Out =>
                  Put (" (conv_out)");
            end case;
            case Inst.Sig_Kind is
               when Kind_Signal_No =>
                  null;
               when Kind_Signal_Register =>
                  Put ("(register)");
               when Kind_Signal_Bus =>
                  Put ("(bus)");
            end case;
            Put (" ");
            Put_Sig_Range (Inst.Sig_Indexes);
         when Ghdl_Name_Process =>
            Put (" [process]");
         when Ghdl_Name_Package =>
            Put (" [package]");
         when Ghdl_Name_Library
           | Ghdl_Name_Configuration
           | Ghdl_Name_Top =>
            null;
      end case;
      New_Line;
   end Disp_Hierarchy_Instance_Name;

--    --  Disp INST and its brother.
--    procedure Disp_Hierarchy_Chain (Inst : Ghdl_Instance_Name_Acc;
--                                    Cur : String;
--                                    Pre : String);

--    --  Disp INST and its child.
--    procedure Disp_Hierarchy (Inst : Ghdl_Instance_Name_Acc;
--                              Cur : String;
--                              Pre : String)
--    is
--    begin
--       case Inst.Kind is
--          when Ghdl_Name_Entity
--            | Ghdl_Name_Architecture
--            | Ghdl_Name_Block
--            | Ghdl_Name_Generate_Iterative
--            | Ghdl_Name_Generate_Conditional
--            | Ghdl_Name_Component =>
--             if Inst.Child = null then
--                Put (Cur);
--                Disp_Hierarchy_Instance_Name (Inst);
--             else
--                declare
--                   Str : String := Get_String (Inst.Name.all);
--                begin
--                   Disp_Hierarchy_Chain (Inst.Child,
--                                         Cur & Str,
--                                         Pre & (1 .. Str'Length => ' '));
--                end;
--             end if;
--          when Ghdl_Name_Signal =>
--             Put (Cur);
--             Disp_Hierarchy_Instance_Name (Inst);
--          when Ghdl_Name_Library
--            | Ghdl_Name_Package
--            | Ghdl_Name_Process
--            | Ghdl_Name_Configuration
--            | Ghdl_Name_Top =>
--             null;
--       end case;
--    end Disp_Hierarchy;

--    --  Disp INST and its brother.
--    procedure Disp_Hierarchy_Chain (Inst : Ghdl_Instance_Name_Acc;
--                                    Cur : String;
--                                    Pre : String)
--    is
--       N : Ghdl_Instance_Name_Acc;
--    begin
--       if Inst.Brother = null then
--          --  Alone.
--          Disp_Hierarchy (Inst, Cur & '-', Pre & ' ');
--       else
--          --  Disp the first one.
--          Disp_Hierarchy (Inst, Cur & "-+-", Pre & " | ");
--          N := Inst.Brother;
--          while N /= null loop
--             if N.Brother /= null then
--                Disp_Hierarchy (N,
--                                Pre & " |-",
--                                Pre & " | ");
--             else
--                Disp_Hierarchy (N,
--                                Pre & " `-",
--                                Pre & "   ");
--             end if;
--             N := N.Brother;
--          end loop;
--       end if;
--    end Disp_Hierarchy_Chain;

--    procedure Disp_Hierarchy is
--    begin
--       Disp_Hierarchy_Chain (Grt.Hierarchy.Top.Child, "", "");
--    end Disp_Hierarchy;

   procedure Disp_Hierarchy_Tree (Inst : Ghdl_Instance_Name_Acc;
                                  Cur : String)
   is
      N : Ghdl_Instance_Name_Acc;
      Br : Ghdl_Instance_Name_Acc;
   begin
      N := Inst;
      while N /= null loop
         Br := N.Brother;
         Put (Cur);
         if Br /= null then
            Put (" +-");
         else
            Put (" `-");
         end if;
         Disp_Hierarchy_Instance_Name (N);
         case N.Kind is
            when Ghdl_Name_Entity
              | Ghdl_Name_Architecture
              | Ghdl_Name_Block
              | Ghdl_Name_Generate_Iterative
              | Ghdl_Name_Generate_Conditional
              | Ghdl_Name_Instance =>
               if N.Child /= null then
                  declare
                     Ncur : String (1 .. Cur'Length + 3);
                  begin
                     Ncur (1 .. Cur'Length) := Cur;
                     Ncur (Cur'Length + 1) := ' ';
                     if Br /= null then
                        Ncur (Cur'Length + 2) := '|';
                     else
                        Ncur (Cur'Length + 2) := ' ';
                     end if;
                     Ncur (Cur'Length + 3) := ' ';
                     Disp_Hierarchy_Tree (N.Child, Ncur);
                  end;
               end if;
            when Ghdl_Name_Signal =>
               null;
            when Ghdl_Name_Library
              | Ghdl_Name_Package
              | Ghdl_Name_Process
              | Ghdl_Name_Configuration
              | Ghdl_Name_Top =>
               null;
         end case;
         N := Br;
      end loop;
   end Disp_Hierarchy_Tree;

   procedure Disp_Hierarchy_Tree
   is
      N : Ghdl_Instance_Name_Acc;
   begin
      N := Grt.Hierarchy.Top.Child;
      while N /= null loop
         Disp_Hierarchy_Instance_Name (N);
         if N.Child /= null then
            Disp_Hierarchy_Tree (N.Child, "");
         end if;
         N := N.Brother;
      end loop;
   end Disp_Hierarchy_Tree;

   procedure Disp_Range (Rptr : Ghdl_Range_Ptr; Desc : Ghdl_Desc_Ptr)
   is
   begin
      case Desc.Kind is
         when Ghdl_Type_I32 =>
            Put_I32 (stdout, Rptr.I32.Left);
            Put (' ');
            Put_Dir (Rptr.I32.Dir);
            Put (' ');
            Put_I32 (stdout, Rptr.I32.Right);
         when others =>
            Internal_Error ("disp_range");
      end case;
   end Disp_Range;

   procedure Disp_Array_Indexes (Desc : Ghdl_Desc_Array; Irange : Address)
   is
      Rptr : Address;
   begin
      Put ("(");
      Rptr := Irange;
      for I in 0 .. Desc.Nbr_Dim - 1 loop
         Disp_Range (To_Ghdl_Range_Ptr (Rptr), Desc.Indexes (I));
         if I /= 0 then
            Put (", ");
            Internal_Error ("disp_array_indexes");
         end if;
      end loop;
      Put (")");
   end Disp_Array_Indexes;

   procedure Disp_Type_Def (Desc : Ghdl_Desc_Ptr)
   is
      Basedesc : Ghdl_Desc_Ptr;
   begin
      case Desc.Kind is
         when Ghdl_Type_B2 =>
            if Desc = To_Ghdl_Desc_Ptr (Std_Standard_Boolean_BTD'Address)
            then
               Put ("boolean");
            elsif Desc = To_Ghdl_Desc_Ptr (Std_Standard_Bit_BTD'Address)
            then
               Put ("bit");
            else
               Put ("B2");
            end if;
         when Ghdl_Type_E8 =>
            if Desc =
              To_Ghdl_Desc_Ptr (Ieee_Std_Logic_1164_Std_Ulogic_BT_BTD'Address)
            then
               Put ("std_ulogic");
            else
               Put ("E8");
            end if;
         when Ghdl_Subtype_E8 =>
            if Desc =
              To_Ghdl_Desc_Ptr (Ieee_Std_Logic_1164_Std_Logic_STD'Address)
            then
               Put ("std_logic");
            else
               Put ("SE8");
            end if;
         when Ghdl_Subtype_Array =>
            Basedesc := Desc.Sub_Arr.Basetype;
            if Basedesc
              = To_Ghdl_Desc_Ptr (Std_Standard_Bit_Vector_BTD'Address)
            then
               Put ("bit_vector ");
               Disp_Array_Indexes (Basedesc.Arr, Desc.Sub_Arr.Trange);
            elsif Basedesc = To_Ghdl_Desc_Ptr
              (Ieee_Std_Logic_1164_Std_Logic_Vector_BT_BTD'Address)
            then
               Put ("std_logic_vector ");
               Disp_Array_Indexes (Basedesc.Arr, Desc.Sub_Arr.Trange);
            else
               Put ("SUB-ARRAY");
            end if;
         when others =>
            Put ("*unknown*");
            --Put (Ghdl_Type_Kind'Image (Desc.B2.Kind));
      end case;
   end Disp_Type_Def;

   procedure Disp_Signal_Type (Sig : Ghdl_Instance_Name_Acc)
   is
      Desc : Ghdl_Desc_Ptr := Sig.Sig_Type_Desc;
   begin
      Disp_Instance_Name (Sig);
      Put (": ");
      if Desc = null then
         Put_Line ("NULL");
         return;
      end if;
      Disp_Type_Def (Desc);
      New_Line;
   end Disp_Signal_Type;

   procedure Disp_Signals_Type (Instance : Ghdl_Instance_Name_Acc)
   is
      Inst : Ghdl_Instance_Name_Acc;
   begin
      Inst := Instance;
      while Inst /= null loop
         case Inst.Kind is
            when Ghdl_Name_Entity
              | Ghdl_Name_Architecture
              | Ghdl_Name_Block
              | Ghdl_Name_Generate_Iterative
              | Ghdl_Name_Generate_Conditional
              | Ghdl_Name_Instance =>
               Disp_Signals_Type (Inst.Child);
            when Ghdl_Name_Signal =>
               Disp_Signal_Type (Inst);
            when Ghdl_Name_Library
              | Ghdl_Name_Package
              | Ghdl_Name_Process
              | Ghdl_Name_Configuration
              | Ghdl_Name_Top =>
               null;
         end case;
         Inst := Inst.Brother;
      end loop;
   end Disp_Signals_Type;

   procedure Disp_Signals_Type is
   begin
      Disp_Signals_Type (Grt.Hierarchy.Top.Child);
   end Disp_Signals_Type;

   --  Disp transaction TRANS from signal SIG.
   procedure Disp_Transaction (Trans : Transaction_Acc; Sig : Ghdl_Signal_Ptr)
   is
      T : Transaction_Acc;
   begin
      T := Trans;
      loop
         case T.Kind is
            when Trans_Value =>
               Disp_Value (stdout, T.Val, Sig.Type_Desc);
            when Trans_Null =>
               Put ("NULL");
            when Trans_Error =>
               Put ("ERROR");
         end case;
         Put ("@");
         Disp_Time (T.Time);
         T := T.Next;
         exit when T = null;
         Put (", ");
      end loop;
   end Disp_Transaction;

   procedure Disp_Simple_Signal (Idx : Sig_Table_Index; Sources : Boolean)
   is
      Sig : Ghdl_Signal_Ptr;
   begin
      Sig := Sig_Table.Table (Idx);
      if False
        and (Sig.Instance_Name /= null and then Sig.Instance_Name.Name /= null)
      then
         Disp_Instance_Name (Sig.Instance_Name);
      else
         Put_Sig_Index (Idx);
      end if;
      Put (": ");
      case Sig.Mode is
         when Mode_B2 =>
            Put ("b2");
         when Mode_E8 =>
            Put ("e8");
         when Mode_E32 =>
            Put ("e32");
         when Mode_I32 =>
            Put ("i32");
         when Mode_P64 =>
            Put ("p64");
         when Mode_F64 =>
            Put ("f64");
      end case;
      Put (": ");
      if Sig.Active then
         Put ('A');
      else
         Put ('-');
      end if;
      if Sig.Event then
         Put ('E');
      else
         Put ('-');
      end if;
      if Sig.Effective /= null then
         Put ('e');
      else
         Put ('-');
      end if;
      Put (" value=");
      Disp_Value (stdout, Sig.Value, Sig.Type_Desc);
      Put ("; drving=");
      Disp_Value (stdout, Sig.Driving_Value, Sig.Type_Desc);
      if Sources then
         if Sig.Nbr_Ports > 0 then
            Put (';');
            Put_I32 (stdout, Ghdl_I32 (Sig.Nbr_Ports));
            Put (" ports");
         end if;
         if Sig.Nbr_Drivers = 0 then
            Put ("; no driver");
         elsif Sig.Nbr_Drivers = 1 then
            Put ("; trans=");
            Disp_Transaction (Sig.Drivers (0).First_Trans, Sig);
         else
            for I in 0 .. Sig.Nbr_Drivers - 1 loop
               New_Line;
               Put ("   ");
               Disp_Transaction (Sig.Drivers (I).First_Trans, Sig);
            end loop;
         end if;
      end if;
      New_Line;
   end Disp_Simple_Signal;

   procedure Disp_Simple_Signals (Sources : Boolean)
   is
      Instance : Ghdl_Instance_Name_Acc := null;
      Sig : Ghdl_Signal_Ptr;
   begin
      for I in Sig_Table.First .. Sig_Table.Last loop
         Sig := Sig_Table.Table (I);
         if Sig.Instance_Name /= Instance then
            Instance := Sig.Instance_Name;
            if Instance.Name /= null then
               Disp_Instance_Name (Instance);
               Put_Line (":");
            end if;
         end if;
         Disp_Simple_Signal (I, Sources);
      end loop;
   end Disp_Simple_Signals;

   procedure Disp_Now
   is
      use Grt.Processes;
   begin
      Put ("Now is ");
      Disp_Time (Current_Time);
      Put (" +");
      Put_I32 (stdout, Ghdl_I32 (Current_Delta));
      New_Line;
   end Disp_Now;

   procedure Disp_Signals_Order is
   begin
      for I in Propagation.First .. Propagation.Last loop
         case Propagation.Table (I).Kind is
            when Drv_One_Driver =>
               Put ("Drv (drv) ");
            when Eff_One_Driver =>
               Put ("Eff (drv) ");
            when Drv_One_Port =>
               Put ("Drv (prt) ");
            when Eff_One_Port =>
               Put ("Eff (prt) ");
            when Imp_Guard =>
               Put ("Guard ");
            when Imp_Stable =>
               Put ("Stable ");
            when Imp_Quiet =>
               Put ("imp quiet");
            when Imp_Transaction =>
               Put ("imp transaction");
            when Imp_Delayed =>
               Put ("imp delayed");
            when Eff_Actual =>
               Put ("Actual ");
            when Eff_Multiple =>
               Put ("Eff resolved ");
            when Drv_One_Resolved =>
               Put ("Drv 1 resolved ");
            when Eff_One_Resolved =>
               Put ("Eff 1 resolved ");
            when In_Conversion =>
               Put ("In conv ");
            when Out_Conversion =>
               Put ("Out conv ");
            when Drv_Error =>
               Put ("Drv error");
            when Drv_Multiple =>
               Put ("drv multiple");
         end case;
         case Propagation.Table (I).Kind is
            when Drv_One_Driver
              | Eff_One_Driver
              | Drv_One_Port
              | Eff_One_Port
              | Drv_One_Resolved
              | Eff_One_Resolved
              | Imp_Guard
              | Imp_Stable
              | Eff_Actual =>
               Put_Sig_Index (Signal_Ptr_To_Index (Propagation.Table (I).Sig));
               New_Line;
            when Eff_Multiple =>
               Put_Sig_Range (Propagation.Table (I).Resolv.Sig_Range);
               New_Line;
            when In_Conversion
              | Out_Conversion =>
               declare
                  Conv : Sig_Conversion_Acc;
               begin
                  Conv := Propagation.Table (I).Conv;
                  Put_Sig_Range (Conv.Src);
                  Put (" -> ");
                  Put_Sig_Range (Conv.Dest);
                  New_Line;
               end;
            when others =>
               null;
         end case;
      end loop;
   end Disp_Signals_Order;

   procedure Disp_Type_Desc (Stream : FILEs; Desc : Ghdl_Desc_Ptr) is
   begin
      case Desc.B2.Kind is
         when Ghdl_Type_B2 =>
            Put (Stream, "b2");
         when Ghdl_Type_E8 =>
            Put (Stream, "e8");
         when Ghdl_Type_E32 =>
            Put (Stream, "e32");
         when Ghdl_Type_I32 =>
            Put (Stream, "i32");
         when Ghdl_Type_I64 =>
            Put (Stream, "i64");
         when Ghdl_Type_F64 =>
            Put (Stream, "f64");
         when Ghdl_Type_P64 =>
            Put (Stream, "p64");
         when Ghdl_Type_Access =>
            Put (Stream, "access");
         when Ghdl_Type_Array =>
            Put (Stream, "array");
         when Ghdl_Type_Record =>
            Put (Stream, "record");
         when Ghdl_Subtype_Array =>
            Put (Stream, "sub-array");
         when Ghdl_Subtype_E8 =>
            Put (Stream, "sub-e8");
         when Ghdl_Subtype_I32 =>
            Put (Stream, "sub-i32");
         when Ghdl_Subtype_F64 =>
            Put (Stream, "sub-f64");
         when Ghdl_Subtype_P64 =>
            Put (Stream, "sub-p64");
         when Ghdl_Type_Any =>
            Put (Stream, "*any*");
      end case;
      New_Line;
   end Disp_Type_Desc;
end Grt.Disp;
