with System;
with Ada.Unchecked_Conversion;
with Grt.Types; use Grt.Types;

package Grt.Typedesc is
   pragma Preelaborate (Grt.Typedesc);

   type Ghdl_Type_Kind is
     (Ghdl_Type_B2,
      Ghdl_Type_E8,
      Ghdl_Type_E32,
      Ghdl_Type_I32,
      Ghdl_Type_I64,
      Ghdl_Type_F64,
      Ghdl_Type_P64,
      Ghdl_Type_Access,
      Ghdl_Type_Array,
      Ghdl_Type_Record,
      Ghdl_Subtype_Array,
      Ghdl_Subtype_E8,
      Ghdl_Subtype_I32,
      Ghdl_Subtype_F64,
      Ghdl_Subtype_P64,
      Ghdl_Type_Any);

   subtype Ghdl_Subtype_Discrete_Kind is Ghdl_Type_Kind range
     Ghdl_Type_B2 .. Ghdl_Type_I32;

   subtype Ghdl_Subtype_Scalar_Kind is Ghdl_Type_Kind range
     Ghdl_Subtype_E8 .. Ghdl_Subtype_P64;

   type Ghdl_Range_B2 is record
      Left : Ghdl_B2;
      Right : Ghdl_B2;
      Dir : Ghdl_Dir_Type;
      Len : Ghdl_Index_Type;
   end record;

   type Ghdl_Range_E8 is record
      Left : Ghdl_E8;
      Right : Ghdl_E8;
      Dir : Ghdl_Dir_Type;
      Len : Ghdl_Index_Type;
   end record;

   type Ghdl_Range_E32 is record
      Left : Ghdl_E32;
      Right : Ghdl_E32;
      Dir : Ghdl_Dir_Type;
      Len : Ghdl_Index_Type;
   end record;

   type Ghdl_Range_I32 is record
      Left : Ghdl_I32;
      Right : Ghdl_I32;
      Dir : Ghdl_Dir_Type;
      Len : Ghdl_Index_Type;
   end record;

   type Ghdl_Range_Type (K : Ghdl_Subtype_Discrete_Kind := Ghdl_Type_B2)
   is record
      case K is
         when Ghdl_Type_B2 =>
            B2 : Ghdl_Range_B2;
         when Ghdl_Type_E8 =>
            E8 : Ghdl_Range_E8;
         when Ghdl_Type_E32 =>
            E32 : Ghdl_Range_E32;
         when Ghdl_Type_I32 =>
            I32 : Ghdl_Range_I32;
      end case;
   end record;
   pragma Unchecked_Union (Ghdl_Range_Type);

   type Ghdl_Range_Ptr is access all Ghdl_Range_Type;
   type Ghdl_Range_Fat_Array is array (Natural) of Ghdl_Range_Ptr;
   type Ghdl_Range_Arr_Ptr is access all Ghdl_Range_Fat_Array;

   type Ghdl_Desc_Type (K : Ghdl_Type_Kind := Ghdl_Type_Any);
   type Ghdl_Desc_Ptr is access all Ghdl_Desc_Type;
   type Ghdl_Desc_Fat_Array is array (Natural) of Ghdl_Desc_Ptr;
   type Ghdl_Desc_Arr_Ptr is access all Ghdl_Desc_Fat_Array;

   type Ghdl_Desc_B2 is record
      Kind : Ghdl_Type_Kind;
      False_Name : Ghdl_Str_Len_Type;
      True_Name : Ghdl_Str_Len_Type;
   end record;

   type Ghdl_Desc_E8 is record
      Kind : Ghdl_Type_Kind;
      Erange : Ghdl_Range_Ptr;
      Values : Ghdl_Str_Len_Array_Ptr;
   end record;

   type Ghdl_Desc_P64 is record
      Kind : Ghdl_Type_Kind;
      Base_Name : Ghdl_Str_Len_Type;
      Units : Ghdl_Str_Len_Array_Ptr;
   end record;

   type Ghdl_Desc_Array is record
      Kind : Ghdl_Type_Kind;
      El_Desc : Ghdl_Desc_Ptr;
      El_Size : Natural;
      Nbr_Dim : Natural;
      Indexes : Ghdl_Desc_Arr_Ptr;
   end record;

   type Ghdl_Desc_Record is record
      Kind : Ghdl_Type_Kind;
      Nbr_El : Natural;
      Els : Ghdl_Desc_Arr_Ptr;
   end record;

   type Ghdl_Desc_Subarray is record
      Kind : Ghdl_Type_Kind;
      Basetype : Ghdl_Desc_Ptr;
      Trange : System.Address;
   end record;

   type Ghdl_Desc_Subtype_Scalar is record
      Kind : Ghdl_Type_Kind;
      Basetype : Ghdl_Desc_Ptr;
      Srange : Ghdl_Range_Ptr;
   end record;

   type Ghdl_Desc_Type (K : Ghdl_Type_Kind := Ghdl_Type_Any) is record
      case K is
         when Ghdl_Type_Any =>
            Kind : Ghdl_Type_Kind;
         when Ghdl_Type_B2 =>
            B2 : Ghdl_Desc_B2;
         when Ghdl_Type_E8 =>
            E8 : Ghdl_Desc_E8;
         when Ghdl_Type_P64 =>
            P64 : Ghdl_Desc_P64;
         when Ghdl_Subtype_E8 =>
            St : Ghdl_Desc_Subtype_Scalar;
         when Ghdl_Type_Array =>
            Arr : Ghdl_Desc_Array;
         when Ghdl_Subtype_Array =>
            Sub_Arr : Ghdl_Desc_Subarray;
         when Ghdl_Type_Record =>
            Rec : Ghdl_Desc_Record;
         when others =>
            Ot : Ghdl_Desc_B2;
      end case;
   end record;
   pragma Unchecked_Union (Ghdl_Desc_Type);

   function To_Ghdl_Desc_Ptr is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Ghdl_Desc_Ptr);

   function To_Ghdl_Range_Ptr is new Ada.Unchecked_Conversion
     (Source => System.Address, Target => Ghdl_Range_Ptr);

   --  Some well-known type desc.
   --  Note: they are all imported.
   Std_Standard_Boolean_BTD : aliased Ghdl_Desc_B2;
   Std_Standard_Bit_BTD : aliased Ghdl_Desc_B2;
   Std_Standard_Bit_Vector_BTD : aliased Ghdl_Desc_Array;

   --  well-known type desc from IEEE library.
   --  They are weakly imported, in order not to drag the ieee library if not
   --  used.
   Ieee_Std_Logic_1164_Std_Ulogic_BT_BTD : aliased Ghdl_Desc_E8;
   Ieee_Std_Logic_1164_Std_Logic_STD : aliased Ghdl_Desc_Subtype_Scalar;
   Ieee_Std_Logic_1164_Std_Logic_Vector_BT_BTD : aliased Ghdl_Desc_Array;
   Ieee_Std_Logic_1164_Std_Ulogic_Vector_BT_BTD : aliased Ghdl_Desc_Array;

private
   pragma Import (C, Std_Standard_Boolean_BTD,
                  "std__standard__boolean__BTD");
   pragma Import (C, Std_Standard_Bit_BTD,
                  "std__standard__bit__BTD");
   pragma Import (C, Std_Standard_Bit_Vector_BTD,
                  "std__standard__bit_vector__BTD");

   pragma Import (C, Ieee_Std_Logic_1164_Std_Ulogic_BT_BTD,
                  "ieee__std_logic_1164__std_ulogic__BT__BTD");
   pragma Weak_External (Ieee_Std_Logic_1164_Std_Ulogic_BT_BTD);

   pragma Import (C, Ieee_Std_Logic_1164_Std_Logic_STD,
                  "ieee__std_logic_1164__std_logic__STD");
   pragma Weak_External (Ieee_Std_Logic_1164_Std_Logic_STD);

   pragma Import (C, Ieee_Std_Logic_1164_Std_Logic_Vector_BT_BTD,
                  "ieee__std_logic_1164__std_logic_vector__BT__BTD");
   pragma Weak_External (Ieee_Std_Logic_1164_Std_Logic_Vector_BT_BTD);

   pragma Import (C, Ieee_Std_Logic_1164_Std_Ulogic_Vector_BT_BTD,
                  "ieee__std_logic_1164__std_ulogic_vector__BT__BTD");
   pragma Weak_External (Ieee_Std_Logic_1164_Std_Ulogic_Vector_BT_BTD);

end Grt.Typedesc;
