package body Lists is
   function Get_Max_Nbr_Elements (List : List_Type) return Natural;
   pragma Inline (Get_Max_Nbr_Elements);

   procedure Set_Max_Nbr_Elements (List : List_Type; Max : Natural);
   pragma Inline (Set_Max_Nbr_Elements);

   procedure List_Set_Nbr_Elements (List : List_Type; Nbr : Natural);
   pragma Inline (List_Set_Nbr_Elements);

   function Get_Nbr_Elements (List: List_Type) return Natural is
   begin
      return List.Nbr;
   end Get_Nbr_Elements;

   procedure List_Set_Nbr_Elements (List : List_Type; Nbr : Natural) is
   begin
      List.Nbr := Nbr;
   end List_Set_Nbr_Elements;

   function Get_Max_Nbr_Elements (List : List_Type) return Natural is
   begin
      return List.Max;
   end Get_Max_Nbr_Elements;

   procedure Set_Max_Nbr_Elements (List : List_Type; Max : Natural) is
   begin
      List.Max := Max;
   end Set_Max_Nbr_Elements;

   function Get_Nth_Element (List: List_Type; N: Natural)
     return Node_Type
   is
   begin
      if N >= List.Nbr then
         return Null_Node;
      end if;
      return List.Els (N);
   end Get_Nth_Element;

   -- Replace an element selected by position.
   procedure Replace_Nth_Element (List: List_Type; N: Natural; El: Node_Type)
   is
   begin
      if N >= List.Nbr then
         raise Program_Error;
      end if;
      List.Els (N) := El;
   end Replace_Nth_Element;

   -- Be sure an element can be added to LIST.
   -- It doesn't change the number of elements.
   procedure List_Grow (List: List_Type)
   is
      --  Be careful: size in bytes.
      function Alloc (Size : Natural) return Node_Array_Fat_Acc;
      pragma Import (C, Alloc, "malloc");

      function Realloc (Ptr : Node_Array_Fat_Acc; Size : Natural)
        return Node_Array_Fat_Acc;
      pragma Import (C, Realloc, "realloc");

      Tmp : Node_Array_Fat_Acc;
      N : Natural;
   begin
      if List.Nbr < List.Max then
         return;
      end if;
      if List.Max = 0 then
         N := 10;
         Tmp := Alloc (N * Node_Type'Size / System.Storage_Unit);
      else
         N := List.Max * 2;
         Tmp := Realloc (List.Els, N * Node_Type'Size / System.Storage_Unit);
      end if;
      List.Els := Tmp;
      List.Max := N;
   end List_Grow;

   procedure Append_Element (List: List_Type; Element: Node_Type) is
   begin
      if List.Nbr >= List.Max then
         List_Grow (List);
      end if;
      List.Els (List.Nbr) := Element;
      List.Nbr := List.Nbr + 1;
   end Append_Element;

   -- Return the last element of the list, or null.
   function Get_Last_Element (List: List_Type) return Node_Type is
   begin
      if List.Nbr = 0 then
         return Null_Node;
      else
         return List.Els (List.Nbr - 1);
      end if;
   end Get_Last_Element;

   -- Return the first element of the list, or null.
   function Get_First_Element (List: List_Type) return Node_Type is
   begin
      if List.Nbr = 0 then
         return Null_Node;
      else
         return List.Els (0);
      end if;
   end Get_First_Element;

   -- Add (append) an element only if it was not already present in the list.
   procedure Add_Element (List: List_Type; El: Node_Type)
   is
      Nbr : Natural := Get_Nbr_Elements (List);
   begin
      for I in 0 .. Nbr - 1 loop
         if List.Els (I) = El then
            return;
         end if;
      end loop;

      Append_Element (List, El);
   end Add_Element;

   procedure Remove_Nth_Element (List: List_Type; N: Natural)
   is
      Nbr : Natural := Get_Nbr_Elements (List);
   begin
      if N >= Nbr then
         raise Program_Error;
      end if;
      for I in N .. Nbr - 2 loop
         List.Els (I) := List.Els (I + 1);
      end loop;
      List.Nbr := Nbr - 1;
   end Remove_Nth_Element;

   procedure Set_Nbr_Elements (List: List_Type; N: Natural) is
   begin
      if N > Get_Nbr_Elements (List) then
         raise Program_Error;
      end if;
      List_Set_Nbr_Elements (List, N);
   end Set_Nbr_Elements;

   -- Return the position of the last element.
   -- Return -1 if the list is empty.
   function Get_Last_Element_Position (List: List_Type) return Integer is
   begin
      return Get_Nbr_Elements (List) - 1;
   end Get_Last_Element_Position;

   function Get_Nbr_Elements_Safe (List: List_Type) return Natural is
   begin
      if List = Null_List then
         return 0;
      else
         return Get_Nbr_Elements (List);
      end if;
   end Get_Nbr_Elements_Safe;

   -- Empty the list
   procedure Empty_List (List: List_Type) is
   begin
      Set_Nbr_Elements (List, 0);
   end Empty_List;

   function Create_List return List_Type is
   begin
      return new List_Record'(Max => 0, Nbr => 0, Els => null);
   end Create_List;

   procedure Free_List (List : List_Type)
   is
      procedure Free (Ptr : Node_Array_Fat_Acc);
      pragma Import (C, Free, "free");
   begin
      if List /= Null_List and then List.Max > 0 then
         Free (List.Els);
      end if;
   end Free_List;
end Lists;
