with Iirs; use Iirs;

package Sem is
   --  Semantic analysis for chapters 1, 2, 10 (uses clauses) and 11.

   -- Do the semantic analysis of design unit DESIGN_UNIT.
   -- Also add a few node or change some nodes, when for exemple an
   -- identifier is changed into an access to the type.
   procedure Semantic (Design_Unit: Iir_Design_Unit);

   -- Get the current design unit, ie, the parameter of the procedure semantic.
   function Get_Current_Design_Unit return Iir_Design_Unit;

   --  Makes the current design unit depends on UNIT.
   --  UNIT must be either an entit_aspect or a design_unit.
   procedure Add_Dependence (Unit : Iir);

   --  INTER_PARENT contains generics and ports interfaces;
   --  ASSOC_PARENT constains generics and ports map aspects.
   procedure Sem_Generic_Port_Association_Chain
     (Inter_Parent : Iir; Assoc_Parent : Iir);

   --  Check for restrictions in 1.1.1.2
   --  Return FALSE in case of error.
   function Check_Port_Association_Restriction
     (Formal : Iir_Signal_Interface_Declaration;
      Actual : Iir_Signal_Interface_Declaration;
      Assoc : Iir)
     return Boolean;

   --  Return TRUE iff LEFT and RIGHT are (in depth) equal.
   --  This corresponds to conformance rules, LRM 2.7
   function Are_Trees_Equal (Left, Right : Iir) return Boolean;

   --  Semantize an use clause.
   --  This may adds use clauses to the chain.
   procedure Sem_Use_Clause (Clauses : Iir_Use_Clause);

   --  LRM 2.1  Subprogram Declarations.
   --  SUBPRG is either a _specification or a _spec_body.
   function Sem_Subprogram_Declaration (Subprg: Iir) return Iir;

   --  LRM 2.2  Subprogram Bodies.
   procedure Sem_Subprogram_Body (Subprg: Iir);
end Sem;
