with Types; use Types;
with Iirs; use Iirs;
with Sem_Expr; use Sem_Expr;

package Sem_Decls is
   --  The kind of an inteface list.
   type Interface_Kind_Type is (Interface_Generic, Interface_Port,
                                Interface_Procedure, Interface_Function);

   procedure Sem_Interface_Chain (Interface_Chain: Iir;
                                  Interface_Kind : Interface_Kind_Type);

   --  Create predefined operations for DECL.
   procedure Create_Implicit_Operations
     (Decl : Iir; Is_Std_Standard : Boolean := False);

   --  Semantize declarations of PARENT.
   procedure Sem_Declaration_Chain (Parent : Iir);

   -- Check all declarations of DECLS_PARENT are complete
   -- This is used for subprograms in any declarations or deferred
   -- constants in a package declaration.
   -- DECL is the declaration that contains the declaration_list LIST.
   procedure Check_Full_Declaration (Decls_Parent : Iir; Decl: Iir);

   procedure Sem_Iterator (Iterator : Iir_Iterator_Declaration;
                           Staticness : Iir_Staticness);

   --  Semantize actuals of ASSOC_CHAIN.
   --  Check all named associations are after positionnal one.
   --  Return TRUE if no error.
   function Sem_Actual_Of_Association_Chain (Assoc_Chain : Iir) return Boolean;

   --  Semantize association chain ASSOC_CHAIN with interfaces from
   --  INTERFACE_CHAIN.
   --  Return the level of compatibility between the two chains in LEVEL.
   --  If FINISH is true, then ASSOC_CHAIN may be modifies (individual assoc
   --  added), and error messages (if any) are displayed.
   -- If CHECK_MATCH is true, then all interface of INTERFACE_LIST must have
   --   an association or a default value.
   --   This may be false for binding indication.
   -- If IS_PORT is true, then an association may be open even if there is
   --   no default value.
   -- LOC is the association.
   -- Sem_Actual_Of_Association_Chain must have been called before.
   procedure Sem_Association_Chain
     (Interface_Chain : Iir;
      Assoc_Chain: in out Iir;
      Finish: Boolean;
      Check_Match : Boolean;
      Is_Port : Boolean;
      Loc : Iir;
      Level : out Compatibility_Level);

   --  Extract from NAME the named entity whose profile matches SIG.
   function Sem_Signature (Name : Iir; Sig : Iir_Signature) return Iir;

end Sem_Decls;
